!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
PROGRAM dummy_io

  USE mpi

  USE mo_yac_finterface, ONLY: yac_finit, yac_fget_localcomm,      &
       &                       yac_fdef_grid, yac_fdef_points,     &
       &                       yac_fset_global_index,              &
       &                       yac_fset_core_mask,                 &
       &                       yac_fset_mask, yac_fdef_field,      &
       &                       yac_fsearch, yac_ffinalize,         &
       &                       YAC_LOCATION_CELL

  IMPLICIT NONE

  INTEGER, PARAMETER :: pd =  12
  INTEGER, PARAMETER :: rd = 307

  INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(pd,rd) !< double precision
  INTEGER, PARAMETER :: wp = dp                        !< selected working precision

  INTEGER, PARAMETER :: no_of_fields = 8
  INTEGER, PARAMETER :: max_char_length = 132

  INTEGER, PARAMETER :: nbr_cells = 2
  INTEGER, PARAMETER :: nbr_vertices = 4

  REAL(wp), PARAMETER :: YAC_RAD = 0.017453292519943295769_wp ! M_PI / 180

  CHARACTER(LEN=max_char_length) :: field_name(no_of_fields)
  CHARACTER(LEN=max_char_length) :: xml_filename
  CHARACTER(LEN=max_char_length) :: xsd_filename
  CHARACTER(LEN=max_char_length) :: grid_name
  CHARACTER(LEN=max_char_length) :: comp_name

  INTEGER :: i, igrid, ierror

  INTEGER :: comp_id
  INTEGER :: cell_point_ids(2)
  INTEGER :: grid_ids(2)

  INTEGER :: glb_index(nbr_cells)
  INTEGER :: cell_core_mask(nbr_cells)
  INTEGER :: nbr_vertices_per_cell

  REAL(wp), ALLOCATABLE :: buffer_lon(:)
  REAL(wp), ALLOCATABLE :: buffer_lat(:)
  INTEGER, ALLOCATABLE  :: cell_to_vertex(:,:)

  INTEGER, ALLOCATABLE  :: cell_mask(:)
  INTEGER, ALLOCATABLE  :: field_id(:)

  INTEGER :: local_comm, npes, rank

  CALL mpi_init (ierror)

  ! Initialise the coupler
  xml_filename = "dummy_coupling.xml"
  xsd_filename = "coupling.xsd"
  CALL yac_finit ( xml_filename, xsd_filename )

  ! Inform the coupler about what we are
  comp_name = "dummy_io"
  CALL yac_fdef_comp ( comp_name, comp_id )

  CALL yac_fget_localcomm ( local_comm, comp_id )

  CALL mpi_comm_rank ( local_comm, rank, ierror )
  CALL mpi_comm_size ( local_comm, npes, ierror )

  WRITE ( 6 , * ) TRIM(comp_name), " rank ", rank, ": local size is ", npes

  ! Here we define two grids (although numrically identical) to mimick
  ! two different Output server groups

  DO igrid = 1, 2

     IF ( igrid == 1 ) THEN
        grid_name = "ocean_grid"
     ELSE
        grid_name = "atmos_grid"
     ENDIF

     ALLOCATE(buffer_lon(nbr_vertices))
     ALLOCATE(buffer_lat(nbr_vertices))
     ALLOCATE(cell_to_vertex(3,nbr_cells))

     nbr_vertices_per_cell = 3

     ! Define vertices

     !      1
     !     / \
     !    / o \
     !   /     \
     !  2-------3   Eq.
     !   \     /
     !    \ o /
     !     \ /
     !      4

     buffer_lon(1) =  0.0 * YAC_RAD
     buffer_lon(2) = -1.0 * YAC_RAD
     buffer_lon(3) =  1.0 * YAC_RAD
     buffer_lon(4) =  0.0 * YAC_RAD
     buffer_lat(1) =  1.0 * YAC_RAD
     buffer_lat(2) =  0.0 * YAC_RAD
     buffer_lat(3) =  0.0 * YAC_RAD
     buffer_lat(4) = -1.0 * YAC_RAD

     ! Connectivity
     cell_to_vertex(1,1) = 1
     cell_to_vertex(2,1) = 2
     cell_to_vertex(3,1) = 3 ! cell 1
     cell_to_vertex(1,2) = 2
     cell_to_vertex(2,2) = 4
     cell_to_vertex(3,2) = 3 ! cell 2

     ! Definition of an unstructured grid
     CALL yac_fdef_grid (          &
          & grid_name,             &
          & nbr_vertices,          &
          & nbr_cells,             &
          & nbr_vertices_per_cell, &
          & buffer_lon,            &
          & buffer_lat,            &
          & cell_to_vertex,        &
          & grid_ids(igrid) )

     ! Decomposition information

     DO i = 1, nbr_cells
        glb_index(i) = i
        cell_core_mask(i) = 1
     ENDDO

     CALL yac_fset_global_index ( &
          & glb_index,            &
          & YAC_LOCATION_CELL,    &
          & grid_ids(igrid) )
     CALL yac_fset_core_mask ( &
          & cell_core_mask,    &
          & YAC_LOCATION_CELL, &
          & grid_ids(igrid) )

     ! Center points in cells (needed e.g. for nearest neighbour)

     buffer_lon(1) =  0.0 * YAC_RAD
     buffer_lon(2) =  0.0 * YAC_RAD
     buffer_lat(1) =  0.5 * YAC_RAD
     buffer_lat(2) = -0.5 * YAC_RAD

     CALL yac_fdef_points (           &
          & grid_ids(igrid),          &
          & nbr_cells,                &
          & YAC_LOCATION_CELL,        &
          & buffer_lon,               &
          & buffer_lat,               &
          & cell_point_ids(igrid) )

     DEALLOCATE (buffer_lon, buffer_lat, cell_to_vertex)

     ! Mask generation
     ALLOCATE(cell_mask(nbr_cells))
     DO i = 1, nbr_cells
        cell_mask(i) = 1
     ENDDO

     CALL yac_fset_mask (          &
          & cell_mask,             &
          & cell_point_ids(igrid) )

     DEALLOCATE (cell_mask)

  ENDDO

  field_name(1) = "atmos_out1" ! output field
  field_name(2) = "atmos_out2" ! output field
  field_name(3) = "atmos_out3" ! output field
  field_name(4) = "atmos_out4" ! output field

  field_name(5) = "ocean_out1" ! output field
  field_name(6) = "ocean_out2" ! output field
  field_name(7) = "ocean_out3" ! output field
  field_name(8) = "ocean_out4" ! output field

  ALLOCATE(field_id(no_of_fields))

  ! atmosphere output

  DO i = 1, no_of_fields-4
     CALL yac_fdef_field (     &
          & field_name(i),     &
          & comp_id,           &
          & cell_point_ids(2), &
          & 1,                 &
          & field_id(i) )
  ENDDO

  ! ocean output

  DO i = no_of_fields-3, no_of_fields
     CALL yac_fdef_field (     &
          & field_name(i),     &
          & comp_id,           &
          & cell_point_ids(1), &
          & 1,                 &
          & field_id(i) )
  ENDDO

  CALL yac_fsearch ( ierror )

  CALL yac_ffinalize

  CALL mpi_finalize (ierror)

END PROGRAM dummy_io
