package de.mpg.mpimet.yacgui.xml.interpolation;

import de.mpg.mpimet.yacgui.xml.Dom;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Definite interpolation method conservative
 * 
 * @author Maxim Yastremsky
 */

public class ConservativeInterpolation extends Interpolation {
	
	/**
	 * the method of normalisation
	 * {@code FRACAREA, DESTAREA}
	 */
	public enum Normalisation {
		FRACAREA,
		DESTAREA;
		
		public static boolean isValidString(String s) {
			return (s.equals(Dom.Value.FRACAREA) || 
					s.equals(Dom.Value.DESTAREA));
		}
		public static Normalisation objectFromString(String s) {
			if (s.equals(Dom.Value.FRACAREA)) {
				return FRACAREA;
			} else if (s.equals(Dom.Value.DESTAREA)) {
				return DESTAREA;
			}
			return null; 
		}
		
		@Override
		public String toString() {
			if (this == FRACAREA) {
				return Dom.Value.FRACAREA;
			} else if (this == DESTAREA) {
				return Dom.Value.DESTAREA;
			}
			return null;
		}
	}

    private int conservativeOrder;
	private boolean enforcedConservation;
	private boolean partialCoverage;
	private Normalisation normalisation;
	
	public ConservativeInterpolation(int conservativeOrder,
            boolean enforcedConservation, boolean partialCoverage,
			Normalisation normalisation) {

		if (conservativeOrder < 1) {
			throw new IllegalArgumentException("conservativeOrder must be bigger than 0");
		}

		if (conservativeOrder > 2) {
			throw new IllegalArgumentException("conservativeOrder must be less than 3");
		}

		attributesNumber = 4;
		method = InterpolationMethod.CONSERVATIVE;
		this.conservativeOrder = conservativeOrder;
		this.enforcedConservation = enforcedConservation;
		this.partialCoverage = partialCoverage;
		this.normalisation = normalisation;
	}

	/**
	 *
	 * @return integer value of conservative order
	 */
	public int getConservativeOrder() {
		return conservativeOrder;
	}

	/**
	 * 
	 * @return true if enforced conservation is active
	 */
	public boolean getEnforcedConservation() {
		return enforcedConservation;
	}
	
	/**
	 * 
	 * @return true if partial coverage is active
	 */
	public boolean getPartialCoverage() {
		return partialCoverage;
	}
	
	/**
	 * 
	 * @return object of {@link Normalisation} indicating the method of normalisation
	 */
	public Normalisation getNormalisation() {
		return normalisation;
	}
	
	/**
	 * get attribute name by the order of index
	 * @param index the order of attributes on the panel
	 * @return  the name of attribute
	 * 				if 0 return CONSERVATIVE_ORDER, if 1 return ENFORCED_CONSERVATION, if 2 return PARTIAL_COVERAGE, if 3 return NORMALISATION
	 */
	public String getAttributeName(int index) {
		if (index == 0) {
			return Dom.Attribute.CONSERVATIVE_ORDER;
		} else if (index == 1) {
			return Dom.Attribute.ENFORCED_CONSERVATION;
		} else if (index == 2) {
			return Dom.Attribute.PARTIAL_COVERAGE;
		} else if (index == 3) {
			return Dom.Attribute.NORMALISATION;
		}
		return null;
	}
	
	/**
	 * get attribute value by the order of index
	 * @param index the order of attributes on the panel
	 * @return the value of attribute as String
	 * 				if 0 return conservativeOrder, if 1 return enforcedConservation, if 2 return partialCoverage, if 3 return normalisation
	 */
	public String getAttributeValue(int index) {
		if (index == 0) {
			return String.valueOf(conservativeOrder);
		} else if (index == 1) {
			return String.valueOf(enforcedConservation);
		} else if (index == 2) {
			return String.valueOf(partialCoverage);
		} else if (index == 3) {
			return normalisation.toString();
		}
		return null;
	}

}
