/**
 * @file instance.c
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "utils.h"
#include "yac_interface.h"
#include "event.h"
#include "yac_mpi.h"
#include "fields.h"
#include "component.h"
#include "interpolation.h"
#include "ensure_array_size.h"
#include "geometry.h"
#include "dist_grid.h"
#include "interp_grid.h"
#include "interp_weights.h"
#include "interp_method.h"
#include "config_xml.h"

struct yac_instance {

  struct yac_couple_config * couple_config;

  struct yac_component_config * comp_config;

  struct coupling_field ** cpl_fields;
  size_t num_cpl_fields;

  MPI_Comm comm;
};

enum transient_type {
  SRC = 1,
  TGT = 2
};

struct comp_grid_config {
  const char * grid_name;
  const char * comp_name;
};

struct comp_grid_pair_config {
  struct comp_grid_config config[2];
};

struct src_transient_config {
  struct coupling_field * field;
  int id;

  struct yac_interp_stack_config interp_stack;
  const char * weight_file_name;
  struct event * event;
};

struct tgt_transient_config {
  struct coupling_field * field;
  int id;
  struct event * event;
};

struct transient_config {

  int id;

  struct comp_grid_pair_config comp_grid_pair;
  int src_comp_idx;

  struct src_transient_config src_interp_config;
  struct tgt_transient_config tgt_interp_config;

  enum interp_weights_reorder_type reorder_type;

  size_t collection_size;
};

static struct yac_basic_grid *
  get_basic_grid(const char * grid_name, int global_component_id,
                 size_t num_fields, struct coupling_field ** coupling_fields) {

  for (size_t i = 0; i < num_fields; ++i) {

    struct coupling_field * field = coupling_fields[i];

    if (global_component_id != yac_get_coupling_field_global_comp_id(field))
      continue;

    struct yac_basic_grid * grid = yac_coupling_field_get_basic_grid(field);

    if (strcmp(grid_name, yac_basic_grid_get_name(grid)) == 0)
      return grid;
  }

  return NULL;
}

static struct component *
get_local_component(
  int global_comp_id, size_t num_components, struct component ** components) {

  for (size_t i = 0; i < num_components; ++i)
    if (yac_component_get_global_id(components[i]) == global_comp_id)
      return components[i];

  return NULL;
}

static
int compare_comp_grid_config(const void * a, const void * b) {

  struct comp_grid_config * a_ = (struct comp_grid_config *)a,
                          * b_ = (struct comp_grid_config *)b;

  int ret;
  if ((ret = strcmp(a_->comp_name, b_->comp_name))) return ret;
  else return strcmp(a_->grid_name, b_->grid_name);
}

static struct coupling_field * get_coupling_field(
  int comp_id, const char * transient_name, const char * grid_name,
  size_t num_fields, struct coupling_field ** coupling_fields) {

  for (size_t i = 0; i < num_fields; ++i) {
    struct coupling_field * curr_field = coupling_fields[i];
    if ((comp_id ==
         yac_get_coupling_field_global_comp_id(curr_field)) &&
         !strcmp(transient_name, yac_get_coupling_field_name(curr_field)) &&
         !strcmp(grid_name,
                 yac_basic_grid_get_name(
                   yac_coupling_field_get_basic_grid(curr_field))))
      return curr_field;
  }

  return NULL;
}

static struct src_transient_config get_src_interp_config(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  struct src_transient_config src_config;

  src_config.interp_stack =
    yac_couple_config_get_interp_stack(
      couple_config, couple_idx, transient_couple_idx);

  // default is 1
  src_config.weight_file_name =
    (yac_couple_config_enforce_write_weight_file(
       couple_config, couple_idx, transient_couple_idx))?
      yac_couple_config_get_weight_file_name(
        couple_config, couple_idx, transient_couple_idx):NULL;

  return src_config;
}

static int compare_transient_config_fields(const void * a, const void * b) {

  struct coupling_field * a_ = (struct coupling_field *)a,
                        * b_ = (struct coupling_field *)b;

  if ((a_ == NULL) || (b_ == NULL)) return (a_ == NULL) - (b_ == NULL);

  size_t num_interp_fields[2] =
    {yac_coupling_field_get_num_interp_fields(a_),
     yac_coupling_field_get_num_interp_fields(b_)};

  // if both transients have different numbers of interpolation fields
  if (num_interp_fields[0] != num_interp_fields[1])
    return (int)num_interp_fields[0] - (int)num_interp_fields[1];

  struct yac_basic_grid * grid[2] =
    {yac_coupling_field_get_basic_grid(a_),
     yac_coupling_field_get_basic_grid(b_)};

  struct interp_field const * interp_fields[2] =
    {yac_coupling_field_get_interp_fields(a_),
     yac_coupling_field_get_interp_fields(b_)};

  for (size_t i = 0; i < num_interp_fields[0]; ++i) {

    int ret;

    enum yac_location location_a = interp_fields[0][i].location;
    enum yac_location location_b = interp_fields[1][i].location;

    if ((ret = (location_a > location_b) -
               (location_a < location_b))) return ret;

    coordinate_pointer coordinates_a =
      yac_basic_grid_get_field_coordinates(grid[0], interp_fields[0][i]);
    coordinate_pointer coordinates_b =
      yac_basic_grid_get_field_coordinates(grid[1], interp_fields[1][i]);

    if ((ret = (coordinates_a > coordinates_b) -
               (coordinates_a < coordinates_b))) return ret;

    int const * masks_a =
      yac_basic_grid_get_field_mask(grid[0], interp_fields[0][i]);
    int const * masks_b =
      yac_basic_grid_get_field_mask(grid[1], interp_fields[1][i]);

    if ((ret = (masks_a > masks_b) - (masks_a < masks_b))) return ret;
  }

  return 0;
}

static int compare_transient_config_field_ids(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret = a_->src_interp_config.id - b_->src_interp_config.id;
  if (ret) return ret;
  return a_->tgt_interp_config.id - b_->tgt_interp_config.id;
}

static int
compare_transient_config_interp_method(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret;
  if ((ret = (int)(a_->src_interp_config.weight_file_name == NULL) -
             (int)(b_->src_interp_config.weight_file_name == NULL))) return ret;
  if ((a_->src_interp_config.weight_file_name != NULL) &&
      (ret = strcmp(a_->src_interp_config.weight_file_name,
                    b_->src_interp_config.weight_file_name))) return ret;

  return
    yac_interp_stack_config_compare(
      &(a_->src_interp_config.interp_stack),
      &(b_->src_interp_config.interp_stack));
}

static
int compare_transient_config(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  int ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[0].comp_name,
                    b_->comp_grid_pair.config[0].comp_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[1].comp_name,
                    b_->comp_grid_pair.config[1].comp_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[0].grid_name,
                    b_->comp_grid_pair.config[0].grid_name))) return ret;
  if ((ret = strcmp(a_->comp_grid_pair.config[1].grid_name,
                    b_->comp_grid_pair.config[1].grid_name))) return ret;
  if ((ret = a_->src_interp_config.id - b_->src_interp_config.id)) return ret;
  if ((ret = a_->tgt_interp_config.id - b_->tgt_interp_config.id)) return ret;
  if ((ret = compare_transient_config_interp_method(a_, b_))) return ret;
  return (int)(a_->reorder_type) - (int)(b_->reorder_type);
}

struct event * get_event_info(
  struct yac_instance * instance, int couple_idx, int transient_couple_idx,
  enum transient_type transient_type) {

  struct yac_couple_config * couple_config = instance->couple_config;

  enum yac_reduction_type reduction_operation =
    yac_couple_config_get_coupling_period_operation(
      couple_config, couple_idx, transient_couple_idx);

  enum yac_time_unit_type timestep_unit =
    yac_couple_config_get_timestep_unit(couple_config);

  // make sure that the calendar is defined and consistent
  calendarType set_calendar = getCalendarType ();
  calendarType my_calendar  = yac_couple_config_get_calendar(couple_config);

  if ( set_calendar == CALENDAR_NOT_SET ) {
    initCalendar(
      (my_calendar != CALENDAR_NOT_SET)?my_calendar:PROLEPTIC_GREGORIAN);
  } else
    YAC_ASSERT(
      (my_calendar == CALENDAR_NOT_SET) || (set_calendar == my_calendar),
      "ERROR: different calendars in use! Reconfigure model setup.")

  struct event * event = yac_event_new();

  char const * coupling_period =
    yac_couple_config_get_coupling_period(
      couple_config, couple_idx, transient_couple_idx);
  char const * timestep;
  int timelag;
  if ( transient_type == SRC )
    yac_couple_config_get_source_timestep(
      couple_config, couple_idx, transient_couple_idx, &timestep, &timelag);
  else
    yac_couple_config_get_target_timestep(
      couple_config, couple_idx, transient_couple_idx, &timestep, &timelag);

  if ( timestep_unit == C_ISO_FORMAT) {

      yac_event_add(
        event, (void*)timestep, (void*)coupling_period, timestep_unit,
        timelag, reduction_operation,
        yac_couple_config_get_start_datetime(couple_config),
        yac_couple_config_get_end_datetime(couple_config));

  } else {

      char * endptr;
      int coupling_period_int = (int)strtol(coupling_period, &endptr, 10);
      YAC_ASSERT_F(
        *endptr == '\0',
        "ERROR(get_event_info): invalid coupling period string '%s'",
        coupling_period)

      int timestep_int = (int)strtol(timestep, &endptr, 10);
      YAC_ASSERT_F(
        *endptr == '\0',
        "ERROR(get_event_info): invalid timestep string '%s'", timestep)

      yac_event_add(
        event, &timestep_int, &coupling_period_int, timestep_unit,
        timelag, reduction_operation,
        yac_couple_config_get_start_datetime(couple_config),
        yac_couple_config_get_end_datetime(couple_config));
  }

  return event;
}

static void get_transient_configuration(
  struct yac_instance * instance,
  struct transient_config ** transient_configs_, size_t * count) {

  struct yac_couple_config * couple_config = instance->couple_config;
  MPI_Comm comm = instance->comm;
  size_t num_fields = instance->num_cpl_fields;
  struct coupling_field ** coupling_fields = instance->cpl_fields;

  size_t num_couples = yac_couple_config_get_num_couples(couple_config);
  size_t total_num_transients = 0;
  for (size_t couple_idx = 0; couple_idx < num_couples; ++couple_idx)
    total_num_transients +=
      yac_couple_config_get_num_couple_transients(couple_config, couple_idx);

  struct transient_config * transient_configs =
    xmalloc(total_num_transients * sizeof(*transient_configs));

  size_t i = 0;

  // extract component, grid configurations from all transients in
  // the coupling configuration
  for (size_t couple_idx = 0; couple_idx < num_couples; ++couple_idx) {

    size_t curr_num_transients =
      yac_couple_config_get_num_couple_transients(couple_config, couple_idx);

    for (size_t transient_couple_idx = 0; transient_couple_idx < curr_num_transients;
         ++transient_couple_idx, ++i) {

      struct comp_grid_config src_config, tgt_config;

      int src_comp_id, tgt_comp_id;
      yac_couple_config_get_transient_couple_component_ids(
        couple_config, couple_idx, transient_couple_idx, &src_comp_id, &tgt_comp_id);

      src_config.comp_name =
        yac_couple_config_get_component_name_by_id(couple_config, src_comp_id);
      tgt_config.comp_name =
        yac_couple_config_get_component_name_by_id(couple_config, tgt_comp_id);

      yac_couple_config_get_transient_grid_names(
        couple_config, couple_idx, transient_couple_idx,
        &(src_config.grid_name), &(tgt_config.grid_name));

      int src_comp_idx =
        compare_comp_grid_config(&src_config, &tgt_config) > 0;

      const char * transient_name =
        yac_couple_config_get_transient_name(
          couple_config, couple_idx, transient_couple_idx);
      struct coupling_field * src_field =
        get_coupling_field(
          src_comp_id, transient_name, src_config.grid_name,
          num_fields, coupling_fields);
      struct coupling_field * tgt_field =
        get_coupling_field(
          tgt_comp_id, transient_name, tgt_config.grid_name,
          num_fields, coupling_fields);

      size_t collection_size =
        yac_couple_config_get_transient_collection_size(
          couple_config, couple_idx, transient_couple_idx);
      if (src_field != NULL)
        yac_set_coupling_field_collection_size(src_field, collection_size);
      if (tgt_field != NULL)
        yac_set_coupling_field_collection_size(tgt_field, collection_size);

      transient_configs[i].comp_grid_pair.config[src_comp_idx] = src_config;
      transient_configs[i].comp_grid_pair.config[src_comp_idx^1] = tgt_config;
      transient_configs[i].src_comp_idx = src_comp_idx;
      transient_configs[i].src_interp_config =
        get_src_interp_config(couple_config, couple_idx, transient_couple_idx);
      transient_configs[i].src_interp_config.field = src_field;
      transient_configs[i].tgt_interp_config.field = tgt_field;
      transient_configs[i].reorder_type =
        (yac_couple_config_mapping_on_source(
           couple_config, couple_idx, transient_couple_idx))?
        (MAPPING_ON_SRC):(MAPPING_ON_TGT);
      transient_configs[i].src_interp_config.event =
        get_event_info(instance, couple_idx, transient_couple_idx, SRC);
      transient_configs[i].tgt_interp_config.event =
        get_event_info(instance, couple_idx, transient_couple_idx, TGT);
      transient_configs[i].collection_size = collection_size;
    }
  }

  { // determine unique transient field config ids, for each unique configuration
    int * transient_src_field_config_compare_matrix =
      xmalloc((size_t)(total_num_transients * (total_num_transients - 1)) *
              sizeof(*transient_src_field_config_compare_matrix));
    int * transient_tgt_field_config_compare_matrix =
      transient_src_field_config_compare_matrix +
      (size_t)(total_num_transients * (total_num_transients - 1)) / 2;

    for (unsigned i = 0, k = 0; i < total_num_transients; ++i) {
      for (unsigned j = i + 1; j < total_num_transients; ++j, ++k) {
        transient_src_field_config_compare_matrix[k] =
          compare_transient_config_fields(
            transient_configs[i].src_interp_config.field,
            transient_configs[j].src_interp_config.field) != 0;
        transient_tgt_field_config_compare_matrix[k] =
          compare_transient_config_fields(
            transient_configs[i].tgt_interp_config.field,
            transient_configs[j].tgt_interp_config.field) != 0;
      }
    }

    yac_mpi_call(
      MPI_Allreduce(MPI_IN_PLACE, transient_src_field_config_compare_matrix,
                    total_num_transients * (total_num_transients - 1),
                    MPI_INT, MPI_LOR, comm), comm);

    int * transient_src_field_config_ids =
      xcalloc(2 * (size_t)total_num_transients,
              sizeof(*transient_src_field_config_ids));
    int * transient_tgt_field_config_ids =
      transient_src_field_config_ids + (size_t)total_num_transients;

    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_src_field_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_src_field_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_src_field_config_compare_matrix[k])
            transient_src_field_config_ids[j] = (int)id;
        id++;
      }
    }
    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_tgt_field_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_tgt_field_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_tgt_field_config_compare_matrix[k])
            transient_tgt_field_config_ids[j] = (int)id;
        id++;
      }
    }
    free(transient_src_field_config_compare_matrix);

    for (unsigned i = 0; i < total_num_transients; ++i) {
      transient_configs[i].src_interp_config.id =
        transient_src_field_config_ids[i];
      transient_configs[i].tgt_interp_config.id =
        transient_tgt_field_config_ids[i];
    }
    free(transient_src_field_config_ids);
  }

  { // determine unique transient config ids, for each unique configuration
    int * transient_config_compare_matrix =
      xmalloc((size_t)(total_num_transients * (total_num_transients - 1)) / 2 *
              sizeof(*transient_config_compare_matrix));
    for (unsigned i = 0, k = 0; i < total_num_transients; ++i)
      for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
        transient_config_compare_matrix[k] =
          compare_transient_config(
            &(transient_configs[i]), &(transient_configs[j])) != 0;

    yac_mpi_call(
      MPI_Allreduce(MPI_IN_PLACE, transient_config_compare_matrix,
                    (total_num_transients * (total_num_transients - 1)) / 2,
                    MPI_INT, MPI_LOR, comm), comm);

    int * transient_config_ids =
      xcalloc((size_t)total_num_transients, sizeof(*transient_config_ids));

    for (unsigned i = 0, k = 0, id = 1; i < total_num_transients; ++i) {
      if (transient_config_ids[i] != 0) {
        k += total_num_transients - i - 1;
        continue;
      } else {
        transient_config_ids[i] = (int)id;
        for (unsigned j = i + 1; j < total_num_transients; ++j, ++k)
          if (!transient_config_compare_matrix[k])
            transient_config_ids[j] = (int)id;
        id++;
      }
    }
    free(transient_config_compare_matrix);

    for (unsigned i = 0; i < total_num_transients; ++i)
      transient_configs[i].id = transient_config_ids[i];
    free(transient_config_ids);
  }

  *transient_configs_ = transient_configs;
  *count = total_num_transients;
}

static
int compare_transient_config_ids(const void * a, const void * b) {

  struct transient_config * a_ = (struct transient_config *)a,
                          * b_ = (struct transient_config *)b;

  return a_->id - b_->id;
}

static struct interp_weights * generate_interp_weights(
  struct src_transient_config src_interp_config,
  struct interp_grid * interp_grid) {

  struct interp_method ** method_stack =
    yac_interp_stack_config_generate(src_interp_config.interp_stack);
  struct interp_weights * weights =
    yac_interp_method_do_search(method_stack, interp_grid);

  yac_interp_method_delete(method_stack);
  free(method_stack);

  return weights;
}

static struct interp_field get_interp_field_from_coupling_field(
  struct coupling_field * field, MPI_Comm comm) {

  if (field != NULL) {

    YAC_ASSERT(
      yac_coupling_field_get_num_interp_fields(field) == 1,
      "ERROR(get_interp_field_from_coupling_field): unsupported number of "
      "interpolation fields")

    struct interp_field interp_field =
      *(yac_coupling_field_get_interp_fields(field));

    uint64_t data[3] = {(uint64_t)interp_field.location,
                        (uint64_t)interp_field.coordinates_idx,
                        (uint64_t)interp_field.masks_idx};
    yac_mpi_call(
      MPI_Allreduce(
        MPI_IN_PLACE, data, 3, MPI_UINT64_T, MPI_MIN, comm), comm);

    YAC_ASSERT(
      data[0] == (uint64_t)(interp_field.location),
      "ERROR(get_interp_field_from_coupling_field): location mismatch")
    YAC_ASSERT(
      data[1] == (uint64_t)(interp_field.coordinates_idx),
      "ERROR(get_interp_field_from_coupling_field): "
      "coordinates index mismatch")
    YAC_ASSERT(
      data[2] == (uint64_t)(interp_field.masks_idx),
      "ERROR(get_interp_field_from_coupling_field): masks index mismatch")

    return interp_field;

  } else {

    uint64_t data[3] = {(uint64_t)LOC_UNDEFINED, UINT64_MAX, UINT64_MAX};

    yac_mpi_call(
      MPI_Allreduce(
        MPI_IN_PLACE, data, 3, MPI_UINT64_T, MPI_MIN, comm), comm);

    struct interp_field interp_field;
    interp_field.location = (enum yac_location)data[0];
    interp_field.coordinates_idx = (size_t)data[1];
    interp_field.masks_idx = (size_t)data[2];
    return interp_field;
  }
}

static void generate_interpolations(struct yac_instance * instance) {

  struct yac_couple_config * couple_config = instance->couple_config;
  MPI_Comm comm = instance->comm;
  size_t num_components =
    yac_component_config_get_num_components(instance->comp_config);
  struct component ** components =
    yac_component_config_get_components(instance->comp_config);

  // get information about all transients
  struct transient_config * transient_configs;
  size_t transient_count;
  get_transient_configuration(
    instance, &transient_configs, &transient_count);

  // sort transient configurations
  qsort(transient_configs, transient_count, sizeof(*transient_configs),
        compare_transient_config_ids);

  struct dist_grid_pair * dist_grid_pair = NULL;
  struct interp_grid * interp_grid = NULL;
  struct interp_weights * interp_weights = NULL;
  struct interpolation * interp = NULL;
  struct comp_grid_pair_config * prev_comp_grid_pair = NULL;
  MPI_Comm comp_pair_comm = MPI_COMM_NULL;
  int is_active = 0;

  // loop over all transients to build interpolations
  for (size_t i = 0; i < transient_count; ++i) {

    struct transient_config * curr_transient_config = transient_configs + i;
    struct comp_grid_pair_config * curr_comp_grid_pair =
      &(curr_transient_config->comp_grid_pair);

    int build_flag = 0;
    // if the current comp pair is different from the previous one
    // => generate a component pair communicator
    if ((prev_comp_grid_pair == NULL) ||
        (strcmp(prev_comp_grid_pair->config[0].comp_name,
                curr_comp_grid_pair->config[0].comp_name) ||
         strcmp(prev_comp_grid_pair->config[1].comp_name,
                curr_comp_grid_pair->config[1].comp_name))) {

      build_flag = 1;

      if (comp_pair_comm != MPI_COMM_NULL)
        yac_mpi_call(MPI_Comm_free(&comp_pair_comm), comm);

      int comp_is_available[2] = {
        (get_local_component(
           yac_couple_config_get_component_id_by_name(
             couple_config, curr_comp_grid_pair->config[0].comp_name),
           num_components, components) != NULL),
        (get_local_component(
           yac_couple_config_get_component_id_by_name(
             couple_config, curr_comp_grid_pair->config[1].comp_name),
           num_components, components) != NULL)};

      is_active = comp_is_available[0] || comp_is_available[1];

      yac_mpi_call(MPI_Comm_split(comm, is_active, 0,
                                  &comp_pair_comm), comm);

      // determine whether both components are available in the setup
      yac_mpi_call(MPI_Allreduce(MPI_IN_PLACE, comp_is_available, 2,
                                 MPI_INT, MPI_LOR, comp_pair_comm), comm);
      is_active = comp_is_available[0] && comp_is_available[1];
    }

    // if the current configuration differs from the previous one and the local
    // process is involved in this configuration
    if (is_active &&
        (build_flag ||
         strcmp(prev_comp_grid_pair->config[0].grid_name,
                curr_comp_grid_pair->config[0].grid_name) ||
         strcmp(prev_comp_grid_pair->config[1].grid_name,
                curr_comp_grid_pair->config[1].grid_name))) {

      build_flag = 1;

      if (dist_grid_pair != NULL) yac_dist_grid_pair_delete(dist_grid_pair);

      char const * grid_names[2] =
        {curr_comp_grid_pair->config[0].grid_name,
         curr_comp_grid_pair->config[1].grid_name};

      int global_component_ids[2] =
      {yac_couple_config_get_component_id_by_name(
         couple_config, curr_comp_grid_pair->config[0].comp_name),
       yac_couple_config_get_component_id_by_name(
         couple_config, curr_comp_grid_pair->config[1].comp_name)};

      struct yac_basic_grid * basic_grid[2] =
        {get_basic_grid(
           grid_names[0], global_component_ids[0],
           instance->num_cpl_fields, instance->cpl_fields),
         get_basic_grid(
           grid_names[1], global_component_ids[1],
           instance->num_cpl_fields, instance->cpl_fields)};

      dist_grid_pair =
        yac_dist_grid_pair_new(
          basic_grid[0],
          yac_couple_config_get_grid_id_by_name(
            couple_config, curr_comp_grid_pair->config[0].grid_name),
          basic_grid[1],
          yac_couple_config_get_grid_id_by_name(
            couple_config, curr_comp_grid_pair->config[1].grid_name),
          comp_pair_comm);
    }

    // if the current source or target field data differes from the previous
    // one
    if (is_active &&
        (build_flag ||
         compare_transient_config_field_ids(
           curr_transient_config - 1, curr_transient_config))) {

      build_flag = 1;

      size_t num_src_fields = 1;
      struct interp_field const src_fields[1] =
        {get_interp_field_from_coupling_field(
           curr_transient_config->src_interp_config.field, comp_pair_comm)};
      struct interp_field const tgt_field =
        get_interp_field_from_coupling_field(
          curr_transient_config->tgt_interp_config.field, comp_pair_comm);
      int src_comp_idx = transient_configs[i].src_comp_idx;
      int src_grid_id =
        yac_couple_config_get_grid_id_by_name(
          couple_config, curr_comp_grid_pair->config[src_comp_idx].grid_name);
      int tgt_grid_id =
        yac_couple_config_get_grid_id_by_name(
          couple_config,
          curr_comp_grid_pair->config[src_comp_idx^1].grid_name);

      if (interp_grid != NULL) yac_interp_grid_delete(interp_grid);

      // generate interp grid
      interp_grid = yac_interp_grid_new(
        dist_grid_pair, src_grid_id, tgt_grid_id, num_src_fields,
        src_fields, tgt_field);
    }

    // if the current interpolation method stack differes from the previous
    // configuration
    if (is_active &&
        (build_flag ||
         compare_transient_config_interp_method(
           curr_transient_config - 1, curr_transient_config))) {

      build_flag = 1;

      if (interp_weights != NULL) yac_interp_weights_delete(interp_weights);

      // generate interp weights
      interp_weights = generate_interp_weights(
        curr_transient_config->src_interp_config, interp_grid);
    }

    if (is_active &&
        (curr_transient_config->src_interp_config.weight_file_name != NULL)) {

      int src_comp_idx = transient_configs[i].src_comp_idx;

      yac_interp_weights_write_to_file(
        interp_weights,
        curr_transient_config->src_interp_config.weight_file_name,
        curr_comp_grid_pair->config[src_comp_idx].grid_name,
        curr_comp_grid_pair->config[src_comp_idx^1].grid_name, -1);
    }

    // if the current weight reorder method differs from the previous
    // configuration
    if (is_active &&
        (build_flag ||
         (curr_transient_config[-1].reorder_type !=
          curr_transient_config->reorder_type) ||
         (curr_transient_config[-1].collection_size !=
          curr_transient_config->collection_size))) {

      if (interp != NULL) yac_interpolation_delete(interp);

      // generate interpolation
      interp =
        yac_interp_weights_get_interpolation(
          interp_weights, curr_transient_config->reorder_type,
          curr_transient_config->collection_size);
    }

    if (is_active) {

      int is_source = curr_transient_config->src_interp_config.field != NULL;
      int is_target = curr_transient_config->tgt_interp_config.field != NULL;

      struct interpolation * interp_copy = yac_interpolation_copy(interp);

      if (is_source) {
        yac_set_coupling_field_put_op(
          curr_transient_config->src_interp_config.field,
          curr_transient_config->src_interp_config.event, interp_copy);
        yac_interpolation_inc_ref_count(interp_copy);
      } else yac_event_delete(curr_transient_config->src_interp_config.event);

      if (is_target) {
        yac_set_coupling_field_get_op(
          curr_transient_config->tgt_interp_config.field,
          curr_transient_config->tgt_interp_config.event, interp_copy);
        yac_interpolation_inc_ref_count(interp_copy);
      } else yac_event_delete(curr_transient_config->tgt_interp_config.event);

      yac_interpolation_delete(interp_copy);

    } else {
      yac_event_delete(curr_transient_config->src_interp_config.event);
      yac_event_delete(curr_transient_config->tgt_interp_config.event);
    }

    prev_comp_grid_pair = curr_comp_grid_pair;
  }

  yac_interpolation_delete(interp);
  yac_interp_weights_delete(interp_weights);
  yac_interp_grid_delete(interp_grid);
  yac_dist_grid_pair_delete(dist_grid_pair);
  if (comp_pair_comm != MPI_COMM_NULL)
    yac_mpi_call(MPI_Comm_free(&comp_pair_comm), comm);
  free(transient_configs);
}

void yac_setup_instance(struct yac_instance * instance) {

  YAC_ASSERT(
    instance->comp_config != NULL,
    "ERROR(yac_setup_instance): no components have been defined")

  generate_interpolations(instance);
}

MPI_Comm yac_instance_get_pair_rootcomm(
  struct yac_instance * instance,
  char const * comp_name1, char const * comp_name2) {

  return
    yac_component_config_get_pair_rootcomm(
      instance->comp_config, comp_name1, comp_name2);
}

void yac_instance_dummy_new(MPI_Comm comm) {

  MPI_Comm dummy_comm;
  yac_mpi_call(MPI_Comm_split(comm, MPI_UNDEFINED, 0, &dummy_comm), comm);
}

struct yac_instance * yac_instance_new(
  const char * xml_filename, const char * schema_filename, MPI_Comm comm) {

  struct yac_instance * instance = xmalloc(1 * sizeof(*instance));

  instance->couple_config =
    yac_couple_config_xml_new(xml_filename, schema_filename);

  instance->comp_config = NULL;

  instance->cpl_fields = NULL;
  instance->num_cpl_fields = 0;

  yac_mpi_call(MPI_Comm_split(comm, 0, 0, &(instance->comm)), comm);

  return instance;
}

void yac_instance_delete(struct yac_instance * instance) {

  if (instance == NULL) return;

  yac_component_config_delete(instance->comp_config);

  for (size_t i = 0; i < instance->num_cpl_fields; ++i)
    yac_coupling_field_delete(instance->cpl_fields[i]);
  free(instance->cpl_fields);

  yac_couple_config_delete(instance->couple_config);

  yac_mpi_call(MPI_Comm_free(&(instance->comm)), MPI_COMM_WORLD);

  free(instance);
}

struct yac_couple_config * yac_instance_get_couple_config(
  struct yac_instance * instance) {

  return instance->couple_config;
}

void yac_instance_def_datetime(
  struct yac_instance * instance, const char * start_datetime,
  const char * end_datetime ) {

  yac_couple_config_set_datetime(
    instance->couple_config, start_datetime, end_datetime);
}

char * yac_instance_get_start_datetime(struct yac_instance * instance) {
  return (char*)yac_couple_config_get_start_datetime(instance->couple_config);
}

char * yac_instance_get_end_datetime(struct yac_instance * instance) {
  return (char*)yac_couple_config_get_end_datetime(instance->couple_config);
}

void yac_instance_def_components(
  struct yac_instance * instance,
  char const ** comp_names, size_t num_comps) {

  YAC_ASSERT(
    instance->comp_config == NULL,
    "ERROR(yac_instance_def_components): components have already been defined")

  instance->comp_config =
    yac_component_config_new(
      instance->couple_config, comp_names, num_comps, instance->comm);

  int redirect = yac_couple_config_redirect_stdout(instance->couple_config);

  if ( redirect && (num_comps > 0)) {
    MPI_Comm comm =
      yac_component_get_communicator(
        yac_component_config_get_components(instance->comp_config)[0]);
    int rank, size;
    yac_mpi_call(MPI_Comm_rank(comm, &rank), comm);
    yac_mpi_call(MPI_Comm_size(comm, &size), comm);
    int redirect_only_root = (redirect != 1);
    yac_redirstdout(comp_names[0], redirect_only_root, rank, size);
  }
}

struct component * yac_instance_get_component(
  struct yac_instance * instance, size_t component_index) {

  YAC_ASSERT(
    (component_index >= 0) &&
    (yac_component_config_get_num_components(instance->comp_config) >
     component_index),
    "ERROR(yac_instance_get_component): invalid component_index")

  return
    yac_component_config_get_components(instance->comp_config)[component_index];
}

struct coupling_field * yac_instance_add_field(
  struct yac_instance * instance, char const * field_name,
  size_t component_index, struct yac_basic_grid * grid,
  struct interp_field * interp_fields, size_t num_interp_fields) {

  struct yac_couple_config * couple_config = instance->couple_config;

  YAC_ASSERT_F(
    yac_couple_config_contains_grid_name(
      couple_config, yac_basic_grid_get_name(grid)),
    "ERROR(yac_instance_add_field): grid \"%s\" not in coupling configuration",
    yac_basic_grid_get_name(grid))

  YAC_ASSERT(
    strlen(field_name) <= YAC_MAX_CHARLEN,
    "ERROR(yac_instance_add_field): field name is too long "
    "(maximum is YAC_MAX_CHARLEN)")
  struct component * component =
    yac_instance_get_component(instance, component_index);

  int global_transient_id =
    yac_couple_config_get_transient_id_by_name(
      couple_config, field_name);

  YAC_ASSERT_F(
    global_transient_id != -1,
    "ERROR(yac_instance_add_field): field \"%s\" not in coupling configuration",
    field_name)
  YAC_ASSERT_F(
    yac_couple_config_transient_config_is_valid(
      couple_config,
      yac_component_get_global_id(component),
      yac_couple_config_get_grid_id_by_name(
        couple_config, yac_basic_grid_get_name(grid)),
      global_transient_id),
    "ERROR(yac_instance_add_field): the component-grid-field combination "
    "\"%s\"-\"%s\"-\"%s\" not in coupling configuration",
    yac_component_get_name(component),
    yac_basic_grid_get_name(grid), field_name)

  struct coupling_field * cpl_field =
    yac_coupling_field_new(
      field_name, component, grid, interp_fields, num_interp_fields);

  instance->cpl_fields =
    xrealloc(
      instance->cpl_fields,
      (instance->num_cpl_fields + 1) * sizeof(*(instance->cpl_fields)));
  instance->cpl_fields[instance->num_cpl_fields] = cpl_field;
  instance->num_cpl_fields++;

  return cpl_field;
}

void yac_instance_get_fields(
  struct yac_instance * instance,
  struct coupling_field *** fields, size_t * num_fields) {

  *fields = instance->cpl_fields;
  *num_fields = instance->num_cpl_fields;
}

