!>
!! @file mo_yac_finterface.f90
!! @brief Implementation of the YAC Fortran user interface description
!!
!! @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
!!                                 Rene Redler <rene.redler@mpimet.mpg.de>
!!                                 Thomas Jahns <jahns@dkrz.de>
!!
!! @version 1.0
!! @author Moritz Hanke <hanke@dkrz.de>
!!         Rene Redler <rene.redler@mpimet.mpg.de>
!!         Thomas Jahns <jahns@dkrz.de>
!!

!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
!             Thomas Jahns <jahns@dkrz.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!

!> \example test_dummy_coupling_dble.f90
!! This example simulates a whole model setup with three components (ocean,
!! atmosphere, io). It uses one process for each component.

!> \example test_dummy_coupling_real.f90
!! This example simulates a whole model setup with three components (ocean,
!! atmosphere, io). It uses one process for each component.

!> \example test_dummy_coupling3_dble.f90
!! Fortran version of \ref test_dummy_coupling3_c.c

!> \example test_dummy_coupling3_real.f90
!! Fortran version of \ref test_dummy_coupling3_c.c

!> \example test_dummy_coupling5_dble.f90
!! Fortran version of \ref test_dummy_coupling5_c.c

!> \example test_dummy_coupling5_real.f90
!! Fortran version of \ref test_dummy_coupling5_c.c

!> \example test_dummy_coupling6_dble.f90
!! Fortran version of \ref test_dummy_coupling6_c.c

!> \example test_dummy_coupling6_real.f90
!! Fortran version of \ref test_dummy_coupling6_c.c

!> \example test_restart_dble.f90
!! Fortran version of \ref test_restart_c.c

#ifdef HAVE_CONFIG_H
! Get the definition of the 'YAC_MPI_FINT_FC_KIND' macro.
#include "config.h"
#endif

module mo_yac_finterface
  use, intrinsic :: iso_c_binding, only : c_int, c_long, &
                                        & c_long_long, c_short, c_char

  !----------------------------------------------------------------------
  !>
  !!   Constants
  !!
  !----------------------------------------------------------------------

  public

  integer, parameter :: YAC_MAX_CHARLEN = 132
  integer, parameter :: YAC_MPI_FINT_KIND = YAC_MPI_FINT_FC_KIND


  !----------------------------------------------------------------------
  !>
  !!   Point types
  !!
  !----------------------------------------------------------------------

  integer, parameter :: YAC_LOCATION_CELL = 0
  integer, parameter :: YAC_LOCATION_CORNER = 1
  integer, parameter :: YAC_LOCATION_EDGE = 2

  !----------------------------------------------------------------------
  !>
  !!   Action types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: NONE               = 0
   enumerator :: COUPLING           = 1
   enumerator :: RESTART            = 2
   enumerator :: GET_FOR_RESTART    = 3
   enumerator :: PUT_FOR_RESTART    = 4
   enumerator :: GET_FOR_CHECKPOINT = 5
   enumerator :: PUT_FOR_CHECKPOINT = 6
   enumerator :: OUT_OF_BOUND       = 7
  end enum

  !----------------------------------------------------------------------
  !>
  !!   Reduction types (see also event.h)
  !!
  !----------------------------------------------------------------------

  enum, bind(c)
   enumerator :: INSTANT      = 0
   enumerator :: ACCUMULATE   = 1
   enumerator :: TIME_AVERAGE = 2
  end enum

  !----------------------------------------------------------------------
  !>
  !!   pointer data types
  !!
  !----------------------------------------------------------------------

  type yac_r4_ptr
    real, pointer :: p(:)
  end type yac_r4_ptr

  type yac_r8_ptr
    double precision, pointer :: p(:)
  end type yac_r8_ptr

  !----------------------------------------------------------------------
  !>
  !!   Fortran2C interface for yac collective routines
  !!
  !----------------------------------------------------------------------

  !> \example test_init_final.f90
  !! This contains an example on how to use yac_finit and yac_ffinalize.

  !> \example test_init_comm_final.f90
  !! This contains an example on how to use yac_finit_comm and yac_ffinalize.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler initialisation
  !!
  !----------------------------------------------------------------------

  interface yac_finit_comm

     subroutine yac_finit_comm ( xml_filename, schema_filename, world_comm )

       character(len=*), intent(in) :: xml_filename    !< [IN] xml description
       character(len=*), intent(in) :: schema_filename !< [IN] xml schema
       integer, intent(in)          :: world_comm      !< [IN] MPI world comm

     end subroutine yac_finit_comm

     subroutine yac_finit_comm_instance ( xml_filename,    &
                                          schema_filename, &
                                          world_comm,      &
                                          yac_instance_id )

       character(len=*), intent(in) :: xml_filename    !< [IN] xml description
       character(len=*), intent(in) :: schema_filename !< [IN] xml schema
       integer, intent(in)          :: world_comm      !< [IN] MPI world comm
       integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

     end subroutine yac_finit_comm_instance

     subroutine yac_finit_comm_dummy ( world_comm )

       integer, intent(in)          :: world_comm      !< [IN] MPI world comm

     end subroutine yac_finit_comm_dummy

  end interface yac_finit_comm

  interface yac_finit

     subroutine yac_finit ( xml_filename, schema_filename )

       character(len=*), intent(in)  :: xml_filename    !< [IN] xml description
       character(len=*), intent(in)  :: schema_filename !< [IN] xml schema

     end subroutine yac_finit

     subroutine yac_finit_instance ( xml_filename,    &
                                     schema_filename, &
                                     yac_instance_id)

       character(len=*), intent(in) :: xml_filename    !< [IN] xml description
       character(len=*), intent(in) :: schema_filename !< [IN] xml schema
       integer, intent(out)         :: yac_instance_id !< [OUT] returned handle to the YAC instance

     end subroutine yac_finit_instance

     subroutine yac_finit_dummy ( )
     end subroutine yac_finit_dummy

  end interface yac_finit

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler cleanup before restart
  !!
  !----------------------------------------------------------------------

  interface yac_fcleanup

     subroutine yac_fcleanup ()
     end subroutine yac_fcleanup

     subroutine yac_fcleanup_instance ( yac_instance_id )

      integer, intent(in) :: yac_instance_id !< [IN] YAC instance identifier

     end subroutine yac_fcleanup_instance

  end interface yac_fcleanup

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the coupler termination
  !!
  !----------------------------------------------------------------------

  interface yac_ffinalize

     subroutine yac_ffinalize ()
     end subroutine yac_ffinalize

     subroutine yac_ffinalize_instance ( yac_instance_id )

      integer, intent(in) :: yac_instance_id !< [IN] YAC instance identifier

     end subroutine

  end interface yac_ffinalize

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting the yac version
  !!
  !----------------------------------------------------------------------

  !> \example test_version.f90
  !! Test for correct setting of YAC version.

  interface
     function yac_fget_version () result (version_string)
     import :: YAC_MAX_CHARLEN
     character(len=YAC_MAX_CHARLEN) :: version_string
     end function yac_fget_version
  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the component definition
  !!
  !----------------------------------------------------------------------

  !> \example test_def_comps.f90
  !! This contains an example on how to use yac_fdef_comp.

  interface yac_fdef_comp

     subroutine yac_fdef_comp ( comp_name, comp_id )

       character(len=*), intent(in) :: comp_name !< [IN]   comp name
       integer, intent(out)         :: comp_id   !< [OUT]  component identifier

     end subroutine yac_fdef_comp

     subroutine yac_fdef_comp_instance ( yac_instance_id, comp_name, comp_id )

       integer, intent(in)          :: yac_instance_id !< [IN] YAC instance identifier
       character(len=*), intent(in) :: comp_name       !< [IN]   comp name
       integer, intent(out)         :: comp_id         !< [OUT]  component identifier

     end subroutine yac_fdef_comp_instance

  end interface yac_fdef_comp

  interface yac_fdef_comps

     subroutine yac_fdef_comps ( comp_names, num_comps, comp_ids )

        use, intrinsic :: iso_c_binding, only : c_char
        import :: YAC_MAX_CHARLEN

        integer, intent(in)   :: num_comps
        character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                                 comp_names(num_comps)
        integer, intent(out)  :: comp_ids(num_comps)

     end subroutine yac_fdef_comps

     subroutine yac_fdef_comps_instance ( yac_instance_id, &
                                          comp_names,      &
                                          num_comps,       &
                                          comp_ids )

        use, intrinsic :: iso_c_binding, only : c_char
        import :: YAC_MAX_CHARLEN

        integer, intent(in)   :: yac_instance_id
        integer, intent(in)   :: num_comps
        character(kind=c_char, len=YAC_MAX_CHARLEN), intent(in) :: &
                                 comp_names(num_comps)
        integer, intent(out)  :: comp_ids(num_comps)

     end subroutine yac_fdef_comps_instance

  end interface yac_fdef_comps

  !> \example test_def_datetime.f90
  !! This shows how to use yac_fdef_datetime.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of start- and end datetime
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_datetime

     subroutine yac_fdef_datetime ( start_datetime, end_datetime )

       character(len=*), intent(in), optional :: start_datetime !< [IN] start datetime of job
       character(len=*), intent(in), optional :: end_datetime   !< [IN] end datetime of job

     end subroutine yac_fdef_datetime

     subroutine yac_fdef_datetime_instance ( yac_instance_id, &
                                             start_datetime,  &
                                             end_datetime )

       integer, intent(in)                    :: yac_instance_id !< [IN] YAC instance identifier
       character(len=*), intent(in), optional :: start_datetime  !< [IN] start datetime of job
       character(len=*), intent(in), optional :: end_datetime    !< [IN] end datetime of job

     end subroutine yac_fdef_datetime_instance

  end interface yac_fdef_datetime

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting the start- and end datetime
  !!
  !----------------------------------------------------------------------

  interface yac_fget_start_datetime

     function yac_fget_start_datetime () result (start_datetime_string)
     import :: YAC_MAX_CHARLEN
     character(len=YAC_MAX_CHARLEN) :: start_datetime_string
     end function yac_fget_start_datetime

     function yac_fget_start_datetime_instance (yac_instance_id) &
       result (start_datetime_string)
     import :: YAC_MAX_CHARLEN
     integer, intent(in)            :: yac_instance_id !< [IN] YAC instance identifier
     character(len=YAC_MAX_CHARLEN) :: start_datetime_string
     end function yac_fget_start_datetime_instance

  end interface yac_fget_start_datetime

  interface yac_fget_end_datetime

     function yac_fget_end_datetime () result (end_datetime_string)
     import :: YAC_MAX_CHARLEN
     character(len=YAC_MAX_CHARLEN) :: end_datetime_string
     end function yac_fget_end_datetime

     function yac_fget_end_datetime_instance (yac_instance_id) &
       result (end_datetime_string)
     import :: YAC_MAX_CHARLEN
     integer, intent(in)            :: yac_instance_id !< [IN] YAC instance identifier
     character(len=YAC_MAX_CHARLEN) :: end_datetime_string
     end function yac_fget_end_datetime_instance

  end interface yac_fget_end_datetime

  !> \example test_def_grid.f90
  !! This shows how to use yac_fdef_grid

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of grids
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_grid

     subroutine yac_fdef_grid_nonuniform_r4 ( grid_name,             &
                                              nbr_vertices,          &
                                              nbr_cells,             &
                                              nbr_connections,       &
                                              nbr_vertices_per_cell, &
                                              x_vertices,            &
                                              y_vertices,            &
                                              cell_to_vertex,        &
                                              grid_id)

       character(len=*), intent(in) :: grid_name                        !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell
       real, intent(in)             :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                          !< [OUT] grid ID

     end subroutine yac_fdef_grid_nonuniform_r4

     subroutine yac_fdef_grid_nonuniform_r8 ( grid_name,             &
                                              nbr_vertices,          &
                                              nbr_cells,             &
                                              nbr_connections,       &
                                              nbr_vertices_per_cell, &
                                              x_vertices,            &
                                              y_vertices,            &
                                              cell_to_vertex,        &
                                              grid_id)

       character(len=*), intent(in) :: grid_name                        !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                     !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                        !< [IN]  number of cells
       integer, intent(in)          :: nbr_connections                  !< [IN]  total size of cell_to_vertex
       integer, intent(in)          :: nbr_vertices_per_cell(nbr_cells) !< [IN]  number of vertices for each cell
       double precision, intent(in) :: x_vertices(nbr_vertices)         !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)         !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex(nbr_connections)  !< [IN] connectivity between vertices and cells\n
                                                                        !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                          !< [OUT] grid ID

     end subroutine yac_fdef_grid_nonuniform_r8

     subroutine yac_fdef_grid_unstruct_r4 ( grid_name,             &
                                            nbr_vertices,          &
                                            nbr_cells,             &
                                            nbr_vertices_per_cell, &
                                            x_vertices,            &
                                            y_vertices,            &
                                            cell_to_vertex,        &
                                            grid_id)

       character(len=*), intent(in) :: grid_name                          !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
       integer, intent(in)          :: nbr_vertices_per_cell              !< [IN]  number of vertices for each cell
       real, intent(in)             :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex( &
                                         nbr_vertices_per_cell,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                          !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                            !< [OUT] grid ID

     end subroutine yac_fdef_grid_unstruct_r4

     subroutine yac_fdef_grid_unstruct_r8 ( grid_name,             &
                                            nbr_vertices,          &
                                            nbr_cells,             &
                                            nbr_vertices_per_cell, &
                                            x_vertices,            &
                                            y_vertices,            &
                                            cell_to_vertex,        &
                                            grid_id)

       character(len=*), intent(in) :: grid_name                          !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices                       !< [IN]  number of vertices
       integer, intent(in)          :: nbr_cells                          !< [IN]  number of cells
       integer, intent(in)          :: nbr_vertices_per_cell              !< [IN]  number of vertices for each cell
       double precision, intent(in) :: x_vertices(nbr_vertices)           !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices)           !< [IN]  latitudes of vertices
       integer, intent(in)          :: cell_to_vertex( &
                                         nbr_vertices_per_cell,nbr_cells) !< [IN] connectivity between vertices and cells\n
                                                                          !!(the vertex indices per cell have to be in clockwise or counterclockwise ordering)
       integer, intent(out)         :: grid_id                            !< [OUT] grid ID

     end subroutine yac_fdef_grid_unstruct_r8

     subroutine yac_fdef_grid_reg2d_r4 ( grid_name,             &
                                         nbr_vertices,          &
                                         cyclic,                &
                                         x_vertices,            &
                                         y_vertices,            &
                                         grid_id)

       character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
       integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension
       real, intent(in)             :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
       real, intent(in)             :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices
       integer, intent(out)         :: grid_id                     !< [OUT] grid ID

     end subroutine yac_fdef_grid_reg2d_r4

     subroutine yac_fdef_grid_reg2d_r8 ( grid_name,             &
                                         nbr_vertices,          &
                                         cyclic,                &
                                         x_vertices,            &
                                         y_vertices,            &
                                         grid_id)

       character(len=*), intent(in) :: grid_name                   !< [IN]  grid name
       integer, intent(in)          :: nbr_vertices(2)             !< [IN]  number of cells in each dimension
       integer, intent(in)          :: cyclic(2)                   !< [IN]  cyclic behavior of cells in each dimension
       double precision, intent(in) :: x_vertices(nbr_vertices(1)) !< [IN]  longitudes of vertices
       double precision, intent(in) :: y_vertices(nbr_vertices(2)) !< [IN]  latitudes of vertices
       integer, intent(out)         :: grid_id                     !< [OUT] grid ID

     end subroutine yac_fdef_grid_reg2d_r8

  end interface yac_fdef_grid

  !> \example test_def_points.f90
  !! This shows how to use yac_fdef_points.

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of points
  !!
  !----------------------------------------------------------------------

  interface yac_fdef_points

     subroutine yac_fdef_points_reg2d_r4 ( grid_id,       &
                                           nbr_points,    &
                                           location,      &
                                           x_points_real, &
                                           y_points_real, &
                                           point_id )

       integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points(2)                !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
       real, intent(in)     :: x_points_real(nbr_points(1)) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points(2)) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r4

     subroutine yac_fdef_points_reg2d_r8 ( grid_id,       &
                                           nbr_points,    &
                                           location,      &
                                           x_points,      &
                                           y_points,      &
                                           point_id )

       integer, intent(in)  :: grid_id                         !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points(2)                   !< [IN]  number of points
       integer, intent(in)  :: location                        !< [IN]  location, one of center/edge/vertex
       double precision, intent(in) :: x_points(nbr_points(1)) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points(2)) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                        !< [OUT] point identifier

     end subroutine yac_fdef_points_reg2d_r8

     subroutine yac_fdef_points_unstruct_r4 ( grid_id,       &
                                              nbr_points,    &
                                              location,      &
                                              x_points_real, &
                                              y_points_real, &
                                              point_id )

       integer, intent(in)  :: grid_id                   !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points                !< [IN]  number of points
       integer, intent(in)  :: location                  !< [IN]  location, one of center/edge/vertex
       real, intent(in)     :: x_points_real(nbr_points) !< [IN]  longitudes of points
       real, intent(in)     :: y_points_real(nbr_points) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                  !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r4

     subroutine yac_fdef_points_unstruct_r8 ( grid_id,       &
                                              nbr_points,    &
                                              location,      &
                                              x_points,      &
                                              y_points,      &
                                              point_id )

       integer, intent(in)  :: grid_id                      !< [IN]  grid identifier
       integer, intent(in)  :: nbr_points                   !< [IN]  number of points
       integer, intent(in)  :: location                     !< [IN]  location, one of center/edge/vertex
       double precision, intent(in) :: x_points(nbr_points) !< [IN]  longitudes of points
       double precision, intent(in) :: y_points(nbr_points) !< [IN]  latitudes of points
       integer, intent(out) :: point_id                     !< [OUT] point identifier

     end subroutine yac_fdef_points_unstruct_r8

  end interface yac_fdef_points

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of grid global ids
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fset_global_index ( global_index, &
                                        location,     &
                                        grid_id )

       integer, intent(in)  :: global_index(*)
       integer, intent(in)  :: location
       integer, intent(in)  :: grid_id

     end subroutine yac_fset_global_index

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of a grid core masks
  !!
  !----------------------------------------------------------------------

  interface yac_fset_core_mask

     subroutine yac_fset_core_imask ( is_core,  &
                                      location, &
                                      grid_id )

       integer, intent(in)  :: is_core(*)
       integer, intent(in)  :: location
       integer, intent(in)  :: grid_id

     end subroutine yac_fset_core_imask

     subroutine yac_fset_core_lmask ( is_core,  &
                                      location, &
                                      grid_id )

       logical, intent(in)  :: is_core(*) !< [IN] mask array
       integer, intent(in)  :: location   !< [IN] location
       integer, intent(in)  :: grid_id    !< [IN] grid identifier

     end subroutine yac_fset_core_lmask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the setting of default pointset masks
  !!
  !----------------------------------------------------------------------

  interface yac_fset_mask

     subroutine yac_fset_imask ( is_valid,   &
                                 points_id )

       integer, intent(in)  :: is_valid(*)
       integer, intent(in)  :: points_id

     end subroutine yac_fset_imask

     subroutine yac_fset_lmask ( is_valid,   &
                                 points_id )

       logical, intent(in)  :: is_valid(*)  !< [IN]  mask array
       integer, intent(in)  :: points_id    !< [IN]  associated points id

     end subroutine yac_fset_lmask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of masks
  !!
  !----------------------------------------------------------------------

  !> \example test_def_mask.f90
  !! This contains an example on how to use yac_fdef_mask.

  interface yac_fdef_mask

     subroutine yac_fdef_imask ( grid_id,    &
                                 nbr_points, &
                                 location,   &
                                 is_valid,   &
                                 mask_id )

       integer, intent(in)  :: grid_id     !< [IN] grid identifier
       integer, intent(in)  :: nbr_points  !< [IN] number of points
       integer, intent(in)  :: location    !< [IN] location, one of center/edge/vertex
       integer, intent(in)  :: is_valid(*) !< [IN] logical mask
                                           !< false, point is masked out
                                           !< true, point is valid
       integer, intent(out) :: mask_id     !< [OUT] mask identifier

     end subroutine yac_fdef_imask

     subroutine yac_fdef_lmask ( grid_id,    &
                                 nbr_points, &
                                 location,   &
                                 is_valid,   &
                                 mask_id )

       integer, intent(in)  :: grid_id     !< [IN] grid identifier
       integer, intent(in)  :: nbr_points  !< [IN] number of points
       integer, intent(in)  :: location    !< [IN] location, one of center/edge/vertex
       logical, intent(in)  :: is_valid(*) !< [IN] logical mask
                                           !< false, point is masked out
                                           !< true, point is valid
       integer, intent(out) :: mask_id     !< [OUT] mask identifier

     end subroutine yac_fdef_lmask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of coupling fields using
  !!   default masks
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_field ( field_name,    &
                                 component_id,  &
                                 point_ids,     &
                                 num_pointsets, &
                                 field_id )

       character(len=*), intent (in) :: field_name    !< [IN] short name of the field
       integer, intent (in)          :: component_id  !< [IN] component identifier
       integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
       integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
       integer, intent (out)         :: field_id      !< [OUT] returned field handle

     end subroutine yac_fdef_field

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for the definition of coupling fields using
  !!   explicit masks
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fdef_field_mask ( field_name,    &
                                      component_id,  &
                                      point_ids,     &
                                      mask_ids,      &
                                      num_pointsets, &
                                      field_id )

       character(len=*), intent (in) :: field_name    !< [IN] short name of the field
       integer, intent (in)          :: component_id  !< [IN] component identifier
       integer, intent (in)          :: point_ids(*)  !< [IN] point identifier
       integer, intent (in)          :: mask_ids(*)   !< [IN] mask identifier
       integer, intent (in)          :: num_pointsets !< [IN] number of pointsets per grid
       integer, intent (out)         :: field_id      !< [OUT] returned field handle

     end subroutine yac_fdef_field_mask

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for sending coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fput

     subroutine yac_fput_r4 ( field_id,         &
                              nbr_hor_points,   &
                              nbr_pointsets,    &
                              collection_size,  &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       real, intent (in)             :: send_field(nbr_hor_points, &
         &                                         nbr_pointsets, &
         &                                         collection_size) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r4

     subroutine yac_fput_r4_ptr ( field_id,         &
          &                       nbr_pointsets,    &
          &                       collection_size,  &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

       import :: yac_r4_ptr

       integer, intent (in)     :: field_id        !< [IN] field identifier
       integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r4_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                                   !< [IN] returned field
       integer, intent (out)    :: info            !< [OUT] returned info
       integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r4_ptr

     subroutine yac_fput_single_pointset_r4 ( field_id,         &
                                              nbr_hor_points,   &
                                              collection_size,  &
                                              send_field,       &
                                              info,             &
                                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       real, intent (in)             :: send_field(nbr_hor_points, collection_size) !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_pointset_r4

     subroutine yac_fput_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              nbr_pointsets,    &
                              send_field,       &
                              info,             &
                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: nbr_pointsets   !< [IN] number of point sets
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (in) :: send_field(nbr_hor_points, nbr_pointsets, collection_size)
                                                        !< [IN] returned field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_r8

     subroutine yac_fput_r8_ptr ( field_id,         &
          &                       nbr_pointsets,    &
          &                       collection_size,  &
          &                       send_field,       &
          &                       info,             &
          &                       ierror )

         import :: yac_r8_ptr

         integer, intent (in)     :: field_id        !< [IN] field identifier
         integer, intent (in)     :: nbr_pointsets   !< [IN] number of point sets
         integer, intent (in)     :: collection_size !< [IN] number of vertical level or bundles
         type(yac_r8_ptr), intent (in) :: send_field(nbr_pointsets, collection_size)
                                                     !< [IN] returned field
         integer, intent (out)    :: info            !< [OUT] returned info
         integer, intent (out)    :: ierror          !< [OUT] returned error

     end subroutine yac_fput_r8_ptr

     subroutine yac_fput_single_pointset_r8 ( field_id,         &
                                              nbr_hor_points,   &
                                              collection_size,  &
                                              send_field,       &
                                              info,             &
                                              ierror )

       integer, intent (in)          :: field_id        !< [IN] field identifier
       integer, intent (in)          :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)          :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (in) :: send_field(nbr_hor_points, collection_size) !< [IN] returned
                                                                                                               !! field
       integer, intent (out)         :: info            !< [OUT] returned info argument
       integer, intent (out)         :: ierror          !< [OUT] returned error handler

     end subroutine yac_fput_single_pointset_r8

  end interface yac_fput

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for receiving coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fget

     subroutine yac_fget_r4 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       real, intent (inout)  :: recv_field(nbr_hor_points, collection_size) !< [INOUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4

     subroutine yac_fget_r4_ptr ( field_id,         &
                                  collection_size,  &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r4_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r4_ptr)      :: recv_field(collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r4_ptr

     subroutine yac_fget_r8 ( field_id,         &
                              nbr_hor_points,   &
                              collection_size,  &
                              recv_field,       &
                              info,             &
                              ierror )

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: nbr_hor_points  !< [IN] number of horizontal points
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       double precision, intent (inout) :: &
          recv_field(nbr_hor_points, collection_size) !< [INOUT] returned
                                                                                                                      !! field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8



     subroutine yac_fget_r8_ptr ( field_id,         &
                                  collection_size,  &
                                  recv_field,       &
                                  info,             &
                                  ierror )

       import :: yac_r8_ptr

       integer, intent (in)  :: field_id        !< [IN] field identifier
       integer, intent (in)  :: collection_size !< [IN] number of vertical level or bundles
       type(yac_r8_ptr)      :: recv_field(collection_size) !< [OUT] returned field
       integer, intent (out) :: info            !< [OUT] returned info
       integer, intent (out) :: ierror          !< [OUT] returned error handler

     end subroutine yac_fget_r8_ptr

  end interface yac_fget

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for exchanging coupling fields
  !!
  !----------------------------------------------------------------------

  interface yac_fexchange

     subroutine yac_fexchange_r4 ( send_field_id,       &
          &                        recv_field_id,       &
          &                        send_nbr_hor_points, &
          &                        send_nbr_pointsets,  &
          &                        recv_nbr_hor_points, &
          &                        collection_size,     &
          &                        send_field,          &
          &                        recv_field,          &
          &                        send_info,           &
          &                        recv_info,           &
          &                        ierror )

       integer, intent (in)  :: send_field_id       !< [IN] field identifier
       integer, intent (in)  :: recv_field_id       !< [IN] field identifier
       integer, intent (in)  :: send_nbr_hor_points !< [IN] number of horizontal send points
       integer, intent (in)  :: send_nbr_pointsets  !< [IN] number of send point sets
       integer, intent (in)  :: recv_nbr_hor_points !< [IN] number of horizontal recv points
       integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
       real, intent (in)     :: send_field(send_nbr_hor_points, &
                                           send_nbr_pointsets,  &
                                           collection_size)
                                                    !< [IN] returned send field
       real, intent (inout)  :: recv_field(recv_nbr_hor_points, &
                                           collection_size)
                                                    !< [INOUT] returned recv field
       integer, intent (out) :: send_info           !< [OUT] returned send info
       integer, intent (out) :: recv_info           !< [OUT] returned recv info
       integer, intent (out) :: ierror              !< [OUT] returned error

     end subroutine yac_fexchange_r4

     subroutine yac_fexchange_r4_ptr ( send_field_id,      &
          &                            recv_field_id,      &
          &                            send_nbr_pointsets, &
          &                            collection_size,    &
          &                            send_field,         &
          &                            recv_field,         &
          &                            send_info,          &
          &                            recv_info,          &
          &                            ierror )

       import :: yac_r4_ptr

       integer, intent (in)  :: send_field_id       !< [IN] send field identifier
       integer, intent (in)  :: recv_field_id       !< [IN] recv field identifier
       integer, intent (in)  :: send_nbr_pointsets  !< [IN] number of send point sets
       integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
       type(yac_r4_ptr), intent (in) ::                        &
                                send_field(send_nbr_pointsets, &
                                           collection_size)
                                                    !< [IN] returned send field
       type(yac_r4_ptr)      :: recv_field(collection_size)
                                                    !< [INOUT] returned recv field
       integer, intent (out) :: send_info           !< [OUT] returned send info
       integer, intent (out) :: recv_info           !< [OUT] returned recv info
       integer, intent (out) :: ierror              !< [OUT] returned error

     end subroutine yac_fexchange_r4_ptr

     subroutine yac_fexchange_single_pointset_r4 ( send_field_id,       &
          &                                        recv_field_id,       &
          &                                        send_nbr_hor_points, &
          &                                        recv_nbr_hor_points, &
          &                                        collection_size,     &
          &                                        send_field,          &
          &                                        recv_field,          &
          &                                        send_info,           &
          &                                        recv_info,           &
          &                                        ierror )

       integer, intent (in)  :: send_field_id        !< [IN] send field identifier
       integer, intent (in)  :: recv_field_id        !< [IN] recv field identifier
       integer, intent (in)  :: send_nbr_hor_points  !< [IN] number of horizontal send points
       integer, intent (in)  :: recv_nbr_hor_points  !< [IN] number of horizontal recv points
       integer, intent (in)  :: collection_size      !< [IN] number of vertical level or bundles
       real, intent (in)     :: send_field(send_nbr_hor_points, &
                                           collection_size)
                                                     !< [IN] returned send field
       real, intent (inout)  :: recv_field(recv_nbr_hor_points, &
                                           collection_size)
                                                     !< [INOUT] returned recv field
       integer, intent (out) :: send_info            !< [OUT] returned send info
       integer, intent (out) :: recv_info            !< [OUT] returned recv info
       integer, intent (out) :: ierror               !< [OUT] returned error

     end subroutine yac_fexchange_single_pointset_r4

     subroutine yac_fexchange_r8 ( send_field_id,       &
          &                        recv_field_id,       &
          &                        send_nbr_hor_points, &
          &                        send_nbr_pointsets,  &
          &                        recv_nbr_hor_points, &
          &                        collection_size,     &
          &                        send_field,          &
          &                        recv_field,          &
          &                        send_info,           &
          &                        recv_info,           &
          &                        ierror )

       integer, intent (in)  :: send_field_id       !< [IN] send field identifier
       integer, intent (in)  :: recv_field_id       !< [IN] recv field identifier
       integer, intent (in)  :: send_nbr_hor_points !< [IN] number of horizontal send points
       integer, intent (in)  :: send_nbr_pointsets  !< [IN] number of send point sets
       integer, intent (in)  :: recv_nbr_hor_points !< [IN] number of horizontal recv points
       integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
       double precision, intent (in) :: &
                                send_field(send_nbr_hor_points, &
                                           send_nbr_pointsets,  &
                                           collection_size)
                                                    !< [IN] returned send field
       double precision, intent (inout):: &
                                recv_field(recv_nbr_hor_points, &
                                           collection_size)
                                                    !< [INOUT] returned recv field
       integer, intent (out) :: send_info           !< [OUT] returned send info
       integer, intent (out) :: recv_info           !< [OUT] returned recv info
       integer, intent (out) :: ierror              !< [OUT] returned error

     end subroutine yac_fexchange_r8

     subroutine yac_fexchange_r8_ptr ( send_field_id,      &
          &                            recv_field_id,      &
          &                            send_nbr_pointsets, &
          &                            collection_size,    &
          &                            send_field,         &
          &                            recv_field,         &
          &                            send_info,          &
          &                            recv_info,          &
          &                            ierror )

       import :: yac_r8_ptr

       integer, intent (in)  :: send_field_id      !< [IN] send field identifier
       integer, intent (in)  :: recv_field_id      !< [IN] recv field identifier
       integer, intent (in)  :: send_nbr_pointsets !< [IN] number of send point sets
       integer, intent (in)  :: collection_size    !< [IN] number of vertical level or bundles
       type(yac_r8_ptr), intent (in) ::                        &
                                send_field(send_nbr_pointsets, &
                                           collection_size)
                                                   !< [IN] returned send field
       type(yac_r8_ptr)      :: recv_field(collection_size)
                                                   !< [OUT] returned recv field
       integer, intent (out) :: send_info          !< [OUT] returned send info
       integer, intent (out) :: recv_info          !< [OUT] returned recv info
       integer, intent (out) :: ierror             !< [OUT] returned error

     end subroutine yac_fexchange_r8_ptr

     subroutine yac_fexchange_single_pointset_r8 ( send_field_id,       &
          &                                        recv_field_id,       &
          &                                        send_nbr_hor_points, &
          &                                        recv_nbr_hor_points, &
          &                                        collection_size,     &
          &                                        send_field,          &
          &                                        recv_field,          &
          &                                        send_info,           &
          &                                        recv_info,           &
          &                                        ierror )

       integer, intent (in)  :: send_field_id       !< [IN] send field identifier
       integer, intent (in)  :: recv_field_id       !< [IN] recv field identifier
       integer, intent (in)  :: send_nbr_hor_points !< [IN] number of horizontal send points
       integer, intent (in)  :: recv_nbr_hor_points !< [IN] number of horizontal recv points
       integer, intent (in)  :: collection_size     !< [IN] number of vertical level or bundles
       double precision, intent (in) ::                         &
                                send_field(send_nbr_hor_points, &
                                           collection_size)
                                                    !< [IN] returned send field
       double precision, intent (inout)::                       &
                                recv_field(recv_nbr_hor_points, &
                                           collection_size)
                                                    !< [INOUT] returned recv field
       integer, intent (out) :: send_info           !< [OUT] returned send info
       integer, intent (out) :: recv_info           !< [OUT] returned recv info
       integer, intent (out) :: ierror              !< [OUT] returned error

     end subroutine yac_fexchange_single_pointset_r8

  end interface yac_fexchange

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for setting and getting the io_proc_step of a component
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in)  :: comp_id      !< [IN] component identifier
       integer, intent (out) :: io_proc_step !< [OUT] stride between ranks of io processes

     end subroutine yac_fget_comp_io_proc_step

     subroutine yac_fset_comp_io_proc_step ( comp_id, io_proc_step )

       integer, intent (in) :: comp_id      !< [IN] component identifier
       integer, intent (in) :: io_proc_step !< [IN] stride between ranks of io processes

     end subroutine yac_fset_comp_io_proc_step

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a local MPI communicator
  !!
  !----------------------------------------------------------------------

  interface

     subroutine yac_fget_localcomm ( local_communicator, comp_id )

       integer, intent (out) :: local_communicator             !< [OUT] MPI communicator
       integer, intent (in)  :: comp_id                        !< [IN]  component ID

     end subroutine yac_fget_localcomm

  end interface

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for getting back a MPI communicator for
  !!   communication between a pair of component root proceses
  !!
  !----------------------------------------------------------------------

  interface yac_fget_pair_rootcomm

     subroutine yac_fget_pair_rootcomm ( local_communicator, &
                                         comp_name1,          &
                                         comp_name2 )

       character(len=*), intent(in) :: comp_name1  !< [IN]  first component name
       character(len=*), intent(in) :: comp_name2  !< [IN]  second component name
       integer, intent (out) :: local_communicator !< [OUT] MPI communicator

     end subroutine yac_fget_pair_rootcomm

     subroutine yac_fget_pair_rootcomm_instance ( yac_instance_id,    &
                                                  local_communicator, &
                                                  comp_name1,         &
                                                  comp_name2 )

       integer, intent(in)          :: yac_instance_id    !< [IN]  YAC instance identifier
       integer, intent(out)         :: local_communicator !< [OUT] component MPI communicator
       character(len=*), intent(in) :: comp_name1         !< [IN]  first component name
       character(len=*), intent(in) :: comp_name2         !< [IN]  second component name

     end subroutine yac_fget_pair_rootcomm_instance

  end interface yac_fget_pair_rootcomm

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking the end of the definition phase
  !!
  !----------------------------------------------------------------------

  interface yac_fsearch

     subroutine yac_fsearch ( ierror )

       integer, intent (out) :: ierror !< [OUT] returned error handler

     end subroutine yac_fsearch

     subroutine yac_fsearch_instance ( yac_instance_id, ierror )

       integer, intent(in)  :: yac_instance_id !< [IN]  YAC instance identifier
       integer, intent(out) :: ierror          !< [OUT] returned error

     end subroutine yac_fsearch_instance

  end interface yac_fsearch

  !----------------------------------------------------------------------
  !>
  !!   Fortran interface for invoking query functions
  !!
  !----------------------------------------------------------------------

  interface yac_fget_nbr_fields

     subroutine yac_fget_nbr_fields ( nbr_fields )

       integer, intent (out) :: nbr_fields !< [OUT] number of coupling fields

     end subroutine yac_fget_nbr_fields

     subroutine yac_fget_nbr_fields_instance ( yac_instance_id, nbr_fields )

       integer, intent(in)   :: yac_instance_id !< [IN]  YAC instance identifier
       integer, intent (out) :: nbr_fields      !< [OUT] number of coupling fields

     end subroutine yac_fget_nbr_fields_instance

  end interface yac_fget_nbr_fields

  ! ---------------------------------------------------------------------

  interface yac_fget_field_ids

     subroutine yac_fget_field_ids ( nbr_fields, field_ids )

       integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
       integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

     end subroutine yac_fget_field_ids

     subroutine yac_fget_field_ids_instance ( yac_instance_id, &
                                              nbr_fields,      &
                                             field_ids )

       integer, intent(in)   :: yac_instance_id       !< [IN]  YAC instance identifier
       integer, intent (in)  :: nbr_fields            !< [IN] number of coupling fields
       integer, intent (out) :: field_ids(nbr_fields) !< [OUT] list of field IDs

     end subroutine yac_fget_field_ids_instance

  end interface yac_fget_field_ids

  ! ---------------------------------------------------------------------

  interface yac_fget_field_id_from_name

     subroutine yac_fget_field_id_from_name ( field_name, field_id )

       character(len=*), intent (in) :: field_name !< [IN]  field name
       integer, intent (out)         :: field_id   !< [OUT] field identifier

     end subroutine yac_fget_field_id_from_name

     subroutine yac_fget_field_id_from_name_instance ( yac_instance_id, &
                                                       field_name,      &
                                                       field_id )

       integer, intent(in)           :: yac_instance_id !< [IN]  YAC instance identifier
       character(len=*), intent (in) :: field_name      !< [IN]  field name
       integer, intent (out)         :: field_id        !< [OUT] field identifier

     end subroutine yac_fget_field_id_from_name_instance

  end interface yac_fget_field_id_from_name

  ! ---------------------------------------------------------------------

  interface yac_fget_field_id_from_field_and_grid_name

     subroutine yac_fget_field_id_from_field_and_grid_name ( &
       field_name, grid_name, field_id )

       character(len=*), intent (in) :: field_name !< [IN]  field name
       character(len=*), intent (in) :: grid_name  !< [IN]  grid name
       integer, intent (out)         :: field_id   !< [OUT] field identifier

     end subroutine yac_fget_field_id_from_field_and_grid_name

     subroutine yac_fget_field_id_from_field_and_grid_name_instance ( &
       yac_instance_id, field_name, grid_name, field_id )

       integer, intent(in)           :: yac_instance_id !< [IN]  YAC instance identifier
       character(len=*), intent (in) :: field_name      !< [IN]  field name
       character(len=*), intent (in) :: grid_name       !< [IN]  grid name
       integer, intent (out)         :: field_id        !< [OUT] field identifier

     end subroutine yac_fget_field_id_from_field_and_grid_name_instance

  end interface yac_fget_field_id_from_field_and_grid_name

  ! ---------------------------------------------------------------------

  interface

     subroutine yac_fget_field_name_from_id ( field_id, field_name )

       integer, intent (in)           :: field_id
       character(len=*), intent (out) :: field_name

     end subroutine yac_fget_field_name_from_id

  end interface

  ! ---------------------------------------------------------------------

  interface yac_fget_role_of_field

    subroutine yac_fget_role_of_field_by_id ( field_id, role )

      integer, intent (in)  :: field_id
      integer, intent (out) :: role

    end subroutine yac_fget_role_of_field_by_id

    ! ....................................................................

    subroutine yac_fget_role_of_field_by_name ( field_name, role )

      character (len=*), intent (in) :: field_name
      integer, intent (out)          :: role

    end subroutine yac_fget_role_of_field_by_name

    subroutine yac_fget_role_of_field_by_name_instance ( yac_instance_id, &
                                                         field_name,      &
                                                         role )

       integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
       character (len=*), intent (in) :: field_name      !< [IN]  field name
       integer, intent (out)          :: role            !< [OUT] role for this field\n
                                                         !!       NOTHING (0) | SOURCE (1) | TARGET (2)

    end subroutine yac_fget_role_of_field_by_name_instance

    ! ....................................................................

    subroutine yac_fget_role_of_field_by_field_and_grid_name ( &
      field_name, grid_name, role )

      character (len=*), intent (in) :: field_name
      character (len=*), intent (in) :: grid_name
      integer, intent (out)          :: role

    end subroutine yac_fget_role_of_field_by_field_and_grid_name

    subroutine yac_fget_role_of_field_by_field_and_grid_name_instance ( &
       yac_instance_id, field_name, grid_name, role )

       integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
       character (len=*), intent (in) :: field_name      !< [IN]  field name
       character (len=*), intent (in) :: grid_name       !< [IN]  grid name
       integer, intent (out)          :: role            !< [OUT] role for this field\n
                                                         !!       NOTHING (0) | SOURCE (1) | TARGET (2)

    end subroutine yac_fget_role_of_field_by_field_and_grid_name_instance

  end interface yac_fget_role_of_field

! ---------------------------------------------------------------------

  interface yac_fget_model_timestep

     subroutine yac_fget_model_timestep_id ( field_id, string )

       integer, intent (in)           :: field_id !< [IN]  field name
       character(len=*), intent (out) :: string   !< [OUT] timestep in iso format

     end subroutine yac_fget_model_timestep_id

     ! ..................................................................

     subroutine yac_fget_model_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name !< [IN]  field name
       character(len=*), intent (out) :: string     !< [OUT] timestep in iso format

     end subroutine yac_fget_model_timestep_name

     ! ..................................................................

     subroutine yac_fget_model_timestep_name_instance ( yac_instance_id, &
                                                        field_name,      &
                                                        string )

       integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
       character(len=*), intent (in)  :: field_name      !< [IN]  field name
       character(len=*), intent (out) :: string          !< [OUT] timestep in iso format

     end subroutine yac_fget_model_timestep_name_instance

  end interface yac_fget_model_timestep

  ! ---------------------------------------------------------------------

  interface yac_fget_coupling_timestep

     subroutine yac_fget_coupling_timestep_id ( field_id, string )

       integer, intent (in)           :: field_id !< [IN]  field identifier
       character(len=*), intent (out) :: string   !< [OUT] timestep in iso format

     end subroutine yac_fget_coupling_timestep_id

     ! ..................................................................

     subroutine yac_fget_coupling_timestep_name ( field_name, string )

       character(len=*), intent (in)  :: field_name !< [IN] field name
       character(len=*), intent (out) :: string     !< [OUT] timestep in iso format

     end subroutine yac_fget_coupling_timestep_name

     ! ..................................................................

     subroutine yac_fget_coupling_timestep_name_instance ( yac_instance_id, &
                                                           field_name,      &
                                                           string )

       integer, intent(in)            :: yac_instance_id !< [IN]  YAC instance identifier
       character(len=*), intent (in)  :: field_name      !< [IN] field name
       character(len=*), intent (out) :: string          !< [OUT] timestep in iso format

     end subroutine yac_fget_coupling_timestep_name_instance

  end interface yac_fget_coupling_timestep

  ! ---------------------------------------------------------------------

  interface yac_redirstdout

     subroutine yac_redirstdout_f ( filestem, parallel, my_pe, npes )

       character(len=*), intent (in) :: filestem
       integer, intent(in) :: parallel
       integer, intent(in) :: my_pe
       integer, intent(in) :: npes

     end subroutine yac_redirstdout_f

  end interface yac_redirstdout

  ! --- ISO_C interface -------------------------------------------------

  !> \example test_abort.f90
  !! This contains an example on how to use yac_abort_message.

  interface

     subroutine yac_abort_message ( text, file, line ) &
           bind ( c, name='yac_abort_message' )

       use, intrinsic :: iso_c_binding, only : c_int, c_char

       character ( kind=c_char ), dimension(*) :: text       !< [IN] error message
       character ( kind=c_char ), dimension(*) :: file       !< [IN] error message
       integer ( kind=c_int ), value           :: line       !< [IN] line number

     end subroutine yac_abort_message

  end interface

end module mo_yac_finterface
