/**
 * @file test_weights2vtk.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <unistd.h> 
#include <stdlib.h>
#include <libgen.h>
#include <string.h>

#include "tests.h"
#include "weight_file_common.h"

int main (int argc, char* argv[]) {

  char const * weights2vtk_exec =
    strcat(dirname(argv[0]), "/../contrib/weights2vtk.x");

  char const * src_grid_name = "test_weights2vtk_src_grid_name";
  char const * tgt_grid_name = "test_weights2vtk_tgt_grid_name";
  char const * weight_file_name = "test_weights2vtk_weight_file_name.nc";
  char const * vtk_file_name = "test_weights2vtk.vtk";

  // test with cubed grid
  {
    // write dummy weight file
    {
      // these links form a message when viewed in 3D
      int src_indices[] = {316,344,346,384,422,413,448,448,469,507,418,83,214};
      int tgt_indices[] = {93,115,70,85,115,108,108,139,169,167,117,50,160};
      double weights[] = {0,1,1,1,1,2,2,2,2,2,3,0,3};
      size_t num_links = 13;
      enum yac_location src_locations[2] = {CORNER, CELL};
      enum yac_location tgt_location = CELL;
      unsigned num_src_fields = 2;
      int num_links_per_field[2] = {11,2};
      int * tgt_id_fixed = NULL;
      size_t num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      size_t num_fixed_values = 0;

      write_weight_file(
        weight_file_name, src_indices, tgt_indices, weights, num_links,
        src_locations, num_src_fields, num_links_per_field, tgt_id_fixed,
        num_fixed_tgt, fixed_values, num_tgt_per_fixed_value,
        num_fixed_values, tgt_location, src_grid_name, tgt_grid_name);
    }

    {
      char cmd[1024];
      sprintf(
        cmd, "%s -S C -T C -s 20 -t 15 -w %s -o %s",
        weights2vtk_exec, weight_file_name, vtk_file_name);
      if (system(cmd)) PUT_ERR("failed to execute");
      unlink(weight_file_name);
      unlink(vtk_file_name);
    }
  }

  // test with curved and unstructed grid, but without links in the weight file
  {
    // write dummy weight file
    {
      int * src_indices = NULL;
      int * tgt_indices = NULL;
      double * weights = NULL;
      size_t num_links = 0;
      enum yac_location src_locations[1] = {CELL};
      enum yac_location tgt_location = CELL;
      unsigned num_src_fields = 1;
      int num_links_per_field[1] = {0};
      int * tgt_id_fixed = NULL;
      size_t num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      size_t num_fixed_values = 0;

      write_weight_file(
        weight_file_name, src_indices, tgt_indices, weights, num_links,
        src_locations, num_src_fields, num_links_per_field, tgt_id_fixed,
        num_fixed_tgt, fixed_values, num_tgt_per_fixed_value,
        num_fixed_values, tgt_location, src_grid_name, tgt_grid_name);
    }

    {
      char cmd[1024];
      sprintf(
        cmd, "%s -S m -T i -s GR30_lsm.nc -t icon_grid_R02B02.nc -w %s -o %s",
        weights2vtk_exec, weight_file_name, vtk_file_name);
      if (system(cmd)) PUT_ERR("failed to execute");
      unlink(weight_file_name);
      unlink(vtk_file_name);
    }
  }

  // test with cubed grid and field data on the edges
  {
    // write dummy weight file (just some dummy data)
    {
      int src_indices[] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
      int tgt_indices[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
      double weights[] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19};
      size_t num_links = 20;
      enum yac_location src_locations[2] = {EDGE};
      enum yac_location tgt_location = EDGE;
      unsigned num_src_fields = 1;
      int num_links_per_field[2] = {20};
      int * tgt_id_fixed = NULL;
      size_t num_fixed_tgt = 0;
      double * fixed_values = NULL;
      int * num_tgt_per_fixed_value = NULL;
      size_t num_fixed_values = 0;

      write_weight_file(
        weight_file_name, src_indices, tgt_indices, weights, num_links,
        src_locations, num_src_fields, num_links_per_field, tgt_id_fixed,
        num_fixed_tgt, fixed_values, num_tgt_per_fixed_value,
        num_fixed_values, tgt_location, src_grid_name, tgt_grid_name);
    }

    {
      char cmd[1024];
      sprintf(
        cmd, "%s -S c -T c -s 5 -t 5 -w %s -o %s",
        weights2vtk_exec, weight_file_name, vtk_file_name);
      if (system(cmd)) PUT_ERR("failed to execute");
      unlink(weight_file_name);
      // unlink(vtk_file_name);
    }
  }

  return TEST_EXIT_CODE;
}
